/**
 * Trust SpamAssassin Headers 0.3.0
 * 
 * by Alex Vallat
 *
 * (based on some code from exampleplugin.cpp)
 */

#using <System.dll>
using namespace System;
using namespace System::Text::RegularExpressions;
using namespace System::Globalization;

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "plugins_services.h"
#include "plugins_spamcheck.h"

/**
 * Receiver identifier - this constant is reserved for future releases of
 * the Plugin SDK.
 */
static const char RECVID_TRUSTSSPAMASSASSINHEADERS[] = "Trust SpamAssassin Headers";

/**
 * Pointer to Spamihilator's SendService function
 * The plugin can send services to Spamihilator through this function
 */
static LPSENDSERVICEFUNC SendService = NULL;

extern "C" __declspec(dllexport) long ReceiveService(SERVICE *service)
{
	switch (service->ulService)
	{
	case SERV_INIT:
		SendService = (LPSENDSERVICEFUNC)service->lParam;
		break;

	case SERV_GETRECV:
		return (long)RECVID_TRUSTSSPAMASSASSINHEADERS;
	}
	return 0;
}

/**
 * Forward declarations
 */
int isSpamHdr(unsigned int id, const MSGHDR *header, double *probability);

/**
 * SPAMCHECKPLUGININFO structure containing informaion about the plugin
 */
SPAMCHECKPLUGININFO spamcheckplugininfo = {
	MAKEVERSION(0,4,0),   //plugin version
	MAKEVERSION(0,9,8),   //Spamihilator version needed (0.9.8)
	"SpamAssassin Headers",     //plugin name
	"Alex Vallat",      //author's name
	"Applies scoring found in headers created by SpamAssassin", //plugin description

	0,					  //init()
	0,                    //free()
	0,					  //config()

	isSpamHdr,            //isSpamHdr()
	0,                    //isSpamHtml()
	0,                    //isSpamNoHtml()
	0                     //isSpam()
};


/**
 * Returns a pointer to a SPAMCHECKPLUGININFO structure containing
 * information about the plugin.
 *
 * @return Pointer to a SPAMCHECKPLUGININFO structure
 */
extern "C" __declspec(dllexport) SPAMCHECKPLUGININFO *GetSpamCheckPluginInfo()
{
	return &spamcheckplugininfo;
}

/**
 * Get a named value from the regex match, as a float
*/
float GetValue(Match^ match, String^ name)
{
	return Single::Parse(match->Groups[name]->Value, NumberStyles::AllowDecimalPoint | NumberStyles::AllowLeadingSign, CultureInfo::InvariantCulture);
}

/**
 * Report the tests that were used by SpamAssassin to determine the spam status
*/
void ReportTests(unsigned int id, String^ tests)
{
	if(tests->Length > 255)
		tests = tests->Substring(0, 255); //Limit to 255 characters

	IntPtr marshalledTests = System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(tests);
	
	//Report the tests as the Spamihilator 'spam words' (reason)
	SERVICE fsw;
	fsw.ulSize      = sizeof(fsw);
	fsw.strReceiver = RECV_SPAMIHILATOR;
	fsw.ulService   = SERV_FOUNDSPAMWORDS;
	fsw.wParam      = (short)id;
	fsw.lParam      = (long)(int)marshalledTests;
	if (SendService) SendService(&fsw);
	System::Runtime::InteropServices::Marshal::FreeHGlobal(marshalledTests);
}

/**
 * Check message by headers

 return value:  1 : Non-Spam   (no other filters will be applied)
                0 : unsure     (let other filters decide)
               -1 : Spam!
 */
int isSpamHdr(unsigned int id, const MSGHDR *header, double *probability)
{
	MSGHDRFIELD *temp = header->fieldlist;
	char* spamStatusBody = NULL;
	char* spamReportBody = NULL;
	while (temp)
	{
		if (temp->name && temp->body)
		{
			if(stricmp(temp->name, "X-Spam-Status") == 0)
			{
				spamStatusBody = temp->body;
			}
			else if(stricmp(temp->name, "X-Spam-Report") == 0)
			{
				spamReportBody = temp->body;
			}
		}
		if(spamReportBody != NULL && spamStatusBody != NULL)
			break; //Found all the headers that need finding

		temp = temp->next;
	}

	if(spamStatusBody != NULL)
	{
		// Managed code starts here
		String^ spamStatus = gcnew String(spamStatusBody);
		Match^ match = Regex::Match(spamStatus, "(?<flag>yes|no),\\s*score\\s*=\\s*(?<score>-?\\d+(?:\\.\\d+)?)", RegexOptions::IgnoreCase);
		if(match->Success)
		{				
			float score = GetValue(match, "score");
			bool flag = match->Groups["flag"]->Value->Equals("yes", System::StringComparison::OrdinalIgnoreCase);
			
			//Try and parse the tests out of the status header
			Match^ testsMatch = Regex::Match(spamStatus, "\\btests\\s*=\\s*(?<tests>\\S+)", RegexOptions::IgnoreCase);
			if(testsMatch->Success)
			{
				ReportTests(id, testsMatch->Groups["tests"]->Value);
			}
			else if(spamReportBody != NULL)
			{
				//Couldn't read tests out of X-Spam-Status, try reading them out of X-Spam-Report
				String^ spamReport = gcnew String(spamReportBody);
				Match^ reportMatch = Regex::Match(spamReport, "-{3,}(?:\\s+-?\\d+(?:\\.\\d+)?\\s+(?<test>[^\\s]+).+?)+$");
				if(reportMatch->Success)
				{
					System::Text::StringBuilder^ tests = gcnew System::Text::StringBuilder();
					for each(Capture^ test in reportMatch->Groups["test"]->Captures)
					{
						tests->Append(test->Value);
						tests->Append(",");
					}
					if(tests->Length > 0)
					{
						ReportTests(id, tests->ToString(0, tests->Length - 1)); //-1 to strip off the last ,
					}
				}
			}

			if(score < 0) //Negative score, actively not spam
			{
				if (probability) *probability = 0.0;
				return 1; //Not spam
			}
			else if (!flag) //Positive score, less than requirement. Might be spam, but can't be sure
			{
				if (probability) *probability = 0.5; //TODO: Is there any point in reporting this as a fraction of the required score? Is probablility used by anything?
				return 0; //Unsure
			}
			else //Positive score, greater than requirement. Spam.
			{
				if (probability) *probability = 1.0;
				return -1; //Spam
			}
		}
		return 0; //Couldn't read the header
	}

	return 0; //Couldn't find the X-Spam-Status header
}

